BASE_DIR = None

SECRET_KEY = ''

DEBUG = False

ALLOWED_HOSTS = []

INSTALLED_APPS = [
    'falcon_core',
]

DEFAULT_MEDIA_TYPE = 'falcon.MEDIA_JSON'

REQUEST_TYPE = 'falcon.request.Request'

RESPONSE_TYPE = 'falcon.response.Response'

MIDDLEWARE = []

ROUTER = 'falcon.routing.DefaultRouter'

INDEPENDENT_MIDDLEWARE = False

ROUTER_CONVERTERS = {}

ROUTES = ''

WSGI_APPLICATION = ''

DEPLOY = {}



import os
from datetime import datetime

from falcon_core.config import ENVIRONMENT_VARIABLE
from falcon_core.management.base import BaseCommand
from falcon_core.management.utils import create_package, create_module
from falcon_core.utils import encrypt_sha256


def startproject(name, base_dir):
    base_module_dir = os.path.join(base_dir, name)
    if not os.path.exists(base_dir):
        os.makedirs(base_dir)
    create_module(base_dir, 'manage', f"""import os\n
from falcon_core import management\n
os.environ.setdefault('{ENVIRONMENT_VARIABLE}', '{name}.settings')\n
if __name__ == '__main__':
    management.execute_from_command_line()
""")
    create_package(base_module_dir)
    create_module(base_module_dir, 'settings', f"""import os\n
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))\n
SECRET_KEY = '{encrypt_sha256(os.getlogin() + name + datetime.now().strftime("%Y-%m-%d %H:%M:%S"))}'\n
DEBUG = True\n
ALLOWED_HOSTS = ['localhost']\n
INSTALLED_APPS = [
    'falcon_core',
]\n
DEFAULT_MEDIA_TYPE = 'falcon.MEDIA_JSON'\n
REQUEST_TYPE = 'falcon.request.Request'\n
RESPONSE_TYPE = 'falcon.response.Response'\n
MIDDLEWARE = []\n
ROUTER = 'falcon.routing.DefaultRouter'\n
INDEPENDENT_MIDDLEWARE = False\n
ROUTER_CONVERTERS = {{}}\n
ROUTES = '{name}.routes'\n
WSGI_APPLICATION = '{name}.wsgi.application'\n
DEPLOY = {{}}
""")
    create_module(base_module_dir, 'routes', """from falcon_core.routes import route\n
routes = [
]
""")
    create_module(base_module_dir, 'wsgi', f"""import os\n
from falcon_core.wsgi import get_wsgi_application\n
os.environ.setdefault('{ENVIRONMENT_VARIABLE}', '{name}.settings')\n
application = get_wsgi_application()
""")


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('name', help='Project name.')
        parser.add_argument('folder', help='Project folder.', nargs='?')

    def handle(self, *args, **kwargs):
        name = kwargs.pop('name')
        folder = kwargs.pop('folder')

        current_folder = os.getcwd()

        if folder is None:
            base_dir = os.path.join(current_folder, name)
            startproject(name, base_dir)
        elif folder == '.':
            base_dir = current_folder
            startproject(name, base_dir)
        else:
            base_dir = os.path.join(current_folder, folder)
            startproject(name, base_dir)



class Rule:
    def __init__(self, obj, extras=None):
        self.obj = obj
        self.extras = extras

    def get(self, name, default=None):
        name, *extras = name.split('.')
        attr = getattr(self.obj, name, default)
        for extra in extras:
            attr = getattr(attr, extra, default)
        return attr

    @classmethod
    def setup(cls, obj, extras=None):
        return cls(obj, extras)

    def execute(self):
        raise NotImplementedError


class Rules:
    _rules = {}

    @classmethod
    def get(cls, name, default=None):
        return cls._rules.get(name, default)

    @classmethod
    def register(cls, name, rule=None):
        if rule:
            cls._rules.update(**{name: rule})
        else:
            def wraps_rule(rule):
                cls._rules.update(**{name: rule})
                return rule

            return wraps_rule

    @classmethod
    def execute(cls, obj, key, extras=None):
        rule = cls.get(key)
        if rule:
            return rule.setup(obj, extras).execute()
        raise KeyError('Rule {} does not exists. Existing rules : {}.'.format(key, ', '.join(cls._rules)))


@Rules.register('original')
class OriginalRule(Rule):
    def execute(self):
        return self.obj


@Rules.register('string')
class StrRule(Rule):
    def execute(self):
        if self.obj is not None:
            return str(self.obj)
        return None


@Rules.register('integer')
class IntRule(Rule):
    def execute(self):
        if self.obj is not None:
            return int(self.obj)
        return None


@Rules.register('float')
class FloatRule(Rule):
    def execute(self):
        if self.obj is not None:
            return float(self.obj)
        return None


@Rules.register('boolean')
class BoolRule(Rule):
    def execute(self):
        if self.obj is not None:
            return bool(self.obj)
        return None


@Rules.register('list')
class ListRule(Rule):
    def execute(self):
        if self.obj is not None:
            return list(self.obj)
        return None


@Rules.register('dict')
class DictRule(Rule):
    def execute(self):
        if self.obj is not None:
            return dict(self.obj)
        return None


@Rules.register('object')
class ObjectRule(Rule):
    def execute(self):
        if self.obj and self.extras:
            data = {}
            for attr, rule, *extras in self.extras:
                attr, name = attr.split(':') if ':' in attr else (attr, attr)
                data[name] = Rules.execute(self.get(attr), rule, *extras)
            return data
        return None


@Rules.register('objects')
class ObjectsRule(Rule):
    def execute(self):
        if self.obj and self.extras:
            return [Rules.execute(obj, 'object', self.extras) for obj in self.obj]
        return []


@Rules.register('c_object')
class CreateObjectRule(Rule):
    def execute(self):
        cls, self.extras = self.extras
        if self.extras:
            instance = cls()
            for key, rule, *extras in self.extras:
                require, key = '*' in key, key.replace('*', '')
                key, attr = key.split(':') if ':' in key else (key, key)
                if hasattr(instance, attr) and key in self.obj:
                    setattr(instance, attr, Rules.execute(self.obj.get(key), rule, *extras))
                elif require:
                    return None
            return instance
        return None


@Rules.register('c_objects')
class CreateObjectsRule(Rule):
    def execute(self):
        if self.obj and self.extras:
            return [Rules.execute(obj, 'c_object', self.extras) for obj in self.obj]
        return None



import importlib
import hashlib
import inspect

from falcon_core.config import settings
from falcon_core.rules import Rules


def import_object(s):
    s = s.split('.')
    return getattr(importlib.import_module('.'.join(s[:-1])), s[-1], None)


def load_middleware(s):
    m = import_object(s)
    if inspect.isclass(m):
        return m()
    return m


def flatten(l):
    f = []
    for i in l:
        if isinstance(i, (list, tuple)):
            f.extend(flatten(i))
        else:
            f.append(i)
    return f


def encrypt_sha1(text):
    return hashlib.sha1(str(text).encode()).hexdigest()


def encrypt_sha256(text):
    return hashlib.sha256(str(text).encode()).hexdigest()


def encrypt_sha1_with_secret_key(text):
    return encrypt_sha1(text + settings.SECRET_KEY)


def encrypt_sha256_with_secret_key(text):
    return encrypt_sha256(text + settings.SECRET_KEY)


def dict_from_obj(obj, extras, iterable=False):
    rule = ('object', 'objects')[int(iterable)]
    return Rules.execute(obj, rule, extras)


def obj_from_dict(obj_dict, extras, iterable=False):
    rule = ('c_object', 'c_objects')[int(iterable)]
    return Rules.execute(obj_dict, rule, extras)



import os
import socket
import getpass
import platform
import subprocess

from falcon_core.management.base import BaseCommand
from falcon_core.config import settings


def get_gunicorn_config():
    return """[Unit]
Description=gunicorn daemon
After=network.target

[Service]
User={user}
Group=www-data
WorkingDirectory={path}
ExecStart={gunicorn} --access-logfile - --workers 3 --bind unix:{sock} {name}.wsgi:application

[Install]
WantedBy=multi-user.target"""


def get_nginx_config():
    return """server {{
    listen 80;
    server_name {domains};

    location / {{
        include proxy_params;
        proxy_pass http://unix:{sock};
    }}

    {extras}
}}"""


def paths_exists(*args):
    if not args:
        raise ValueError('Not set arg or args.')
    for arg in args:
        if not os.path.exists(arg):
            return False
    return True


class DeployError(Exception):
    pass


class Command(BaseCommand):
    def _setup(self):
        self.hostname = socket.gethostname()
        self.path = os.getcwd()
        self.user = getpass.getuser()

        self.settings = getattr(settings, 'DEPLOY', {})

        self.gunicorn_config = self.settings.get('GUNICORN_CONFIG', get_gunicorn_config())

        self.nginx_config = self.settings.get('NGINX_CONFIG', get_nginx_config())

        self.nginx_extras = self.settings.get('NGINX_EXTRAS', '')

        self.app = {'path': settings.BASE_DIR, 'name': os.path.basename(settings.BASE_DIR)}

        self.gunicorn = {
            'name': 'gunicorn_{}.service'.format(self.app['name']),
            'file': os.path.join(self.app['name'], 'gunicorn_{}.service'.format(self.app['name'])),
            'enabled_dir': os.path.abspath(os.path.join(os.sep, 'etc', 'systemd', 'system')),
            'enabled_file': os.path.join(os.path.abspath(os.path.join(os.sep, 'etc', 'systemd', 'system')),
                                         'gunicorn_{}.service'.format(self.app['name'])),
            'sock': os.path.join(self.app['path'], '{}.sock'.format(self.app['name'])),
        }

        self.nginx = {
            'name': self.app['name'], 'file': os.path.join(self.app['name'], self.app['name']),
            'available_dir': os.path.abspath(os.path.join(os.sep, 'etc', 'nginx', 'sites-available')),
            'available_file': os.path.join(os.path.abspath(os.path.join(os.sep, 'etc', 'nginx', 'sites-available')),
                                           self.app['name']),
            'enabled_dir': os.path.abspath(os.path.join(os.sep, 'etc', 'nginx', 'sites-enabled')),
            'enabled_file': os.path.join(os.path.abspath(os.path.join(os.sep, 'etc', 'nginx', 'sites-enabled')),
                                         self.app['name'])
        }

        self.domains = ' '.join(settings.ALLOWED_HOSTS)

        if not self.domains or '*' in self.domains:
            raise ValueError('Set correct domains.')

    def add_arguments(self, parser):
        parser.add_argument('--restart', action='store_true', dest='restart', help='Restart gunicorn and nginx.')
        parser.add_argument('--stop', action='store_true', dest='stop', help='Stop gunicorn and nginx.')

    def handle(self, *args, **options):
        if platform.system() == 'Linux':
            self._setup()
            if options.get('restart') and not options.get('stop'):
                self.restart()
            elif not options.get('restart') and options.get('stop'):
                self.stop()
            elif not options.get('restart') and not options.get('stop'):
                self.deploy()
        else:
            print('Only for Linux OS!')

    def command_call(self, command):
        command = 'sudo {}'.format(command)
        print('{}@{}:{}# {}'.format(self.user, self.hostname, self.path, command))
        subprocess.call(command, shell=True)

    def deploy(self):
        self.stop()

        with open(self.gunicorn['file'], 'w') as gc:
            gunicorn_config = self.gunicorn_config.format(
                user=self.user, path=self.app['path'],
                gunicorn=os.path.join(os.environ['VIRTUAL_ENV'], 'bin', 'gunicorn'),
                sock=self.gunicorn['sock'], name=self.app['name'])
            print('{}\n'.format(gunicorn_config))
            gc.write(gunicorn_config)

        self.command_call('mv {} {}'.format(self.gunicorn['file'], self.gunicorn['enabled_dir']))
        self.command_call('systemctl daemon-reload')
        self.command_call('systemctl restart {}'.format(self.gunicorn['name']))
        self.command_call('systemctl start {}'.format(self.gunicorn['name']))
        self.command_call('systemctl enable {}'.format(self.gunicorn['name']))
        self.command_call('systemctl status {}'.format(self.gunicorn['name']))

        with open(self.nginx['file'], 'w') as nc:
            nginx_config = self.nginx_config.format(domains=self.domains, path=self.app['path'],
                                                    sock=self.gunicorn['sock'], extras=self.nginx_extras)
            print('{}\n'.format(nginx_config))
            nc.write(nginx_config)

        self.command_call('mv {} {}'.format(self.nginx['file'], self.nginx['available_dir']))
        self.command_call('ln -s {} {}'.format(self.nginx['available_file'], self.nginx['enabled_dir']))
        self.command_call('nginx -t')
        self.command_call('systemctl restart nginx')
        self.command_call('ufw allow \'Nginx Full\'')

    def restart(self):
        if paths_exists(self.gunicorn['enabled_file'], self.nginx['enabled_file']):
            self.command_call('systemctl daemon-reload')
            self.command_call('systemctl restart {}'.format(self.gunicorn['name']))
            self.command_call('systemctl restart nginx')
        else:
            raise DeployError()

    def stop(self):
        self.command_call('systemctl stop nginx')

        if paths_exists(self.gunicorn['enabled_file']):
            self.command_call('rm {}'.format(self.gunicorn['enabled_file']))

        if paths_exists(self.nginx['available_file']):
            self.command_call('rm {}'.format(self.nginx['available_file']))

        if paths_exists(self.nginx['enabled_file']):
            self.command_call('rm {}'.format(self.nginx['enabled_file']))

        if paths_exists(self.gunicorn['sock']):
            self.command_call('rm {}'.format(self.gunicorn['sock']))

        self.command_call('systemctl start nginx')



from argparse import ArgumentParser


class BaseCommand:
    def create_parser(self):
        parser = ArgumentParser()
        self.add_arguments(parser)
        return parser

    def add_arguments(self, parser):
        pass

    def run_from_argv(self, argv):
        parser = self.create_parser()
        options = parser.parse_args(argv[2:])
        cmd_options = vars(options)
        args = cmd_options.pop('args', ())
        self.execute(*args, **cmd_options)

    def execute(self, *args, **kwargs):
        self.handle(*args, **kwargs)

    def handle(self, *args, **kwargs):
        raise NotImplementedError('Method handle must be implementer.')



import os

from falcon_core.management.base import BaseCommand
from falcon_core.management.utils import create_package, create_module
from falcon_core.config import settings


def startapp(app_name, app_path):
    if not os.path.exists(os.path.join(app_path, 'resources.py')):
        create_module(app_path, 'resources', 'from falcon_core.resources import Resource\n')
    if not os.path.exists(os.path.join(app_path, 'routes.py')):
        create_module(app_path, 'routes', 'from falcon_core.routes import route\n\nroutes = []\n')


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('app_name', action='store')

    def handle(self, *args, **kwargs):
        app_name = kwargs.pop('app_name')
        if '.' in app_name:
            app_name = app_name.split('.')
            app_path = settings.BASE_DIR
            for name in app_name:
                app_path = os.path.join(app_path, name)
                create_package(app_path)
            app_name = app_name[-1]
            startapp(app_name, app_path)
        else:
            app_path = os.path.join(settings.BASE_DIR, app_name)
            create_package(app_path)
            startapp(app_name, app_path)



import os
import importlib

from falcon_core.config import global_settings

ENVIRONMENT_VARIABLE = 'FALCON_CORE_SETTINGS_MODULE'


class LazySettings:
    _wrapped = None

    def __getattr__(self, item):
        if self._wrapped is None:
            self._setup()
        return getattr(self._wrapped, item)

    def _setup(self):
        settings_module = os.environ.get(ENVIRONMENT_VARIABLE)
        if not settings_module:
            raise EnvironmentError(f'Set environment variable {ENVIRONMENT_VARIABLE}.')
        self._wrapped = Settings(settings_module)

    def __repr__(self):
        if self._wrapped is None:
            return '<LazySettings [Unevaluated]>'
        return f'<LazySettings "{self._wrapped.SETTINGS_MODULE}">'


class Settings:
    def __init__(self, settings_module):
        for setting in dir(global_settings):
            if setting.isupper():
                setattr(self, setting, getattr(global_settings, setting))

        self.SETTINGS_MODULE = settings_module
        module = importlib.import_module(self.SETTINGS_MODULE)

        tuple_settings = (
            'ALLOWED_HOSTS',
            'INSTALLED_APPS',
            'MIDDLEWARE',
        )

        self._explicit_settings = set()

        for setting in dir(module):
            if setting.isupper():
                setting_value = getattr(module, setting)

                if setting in tuple_settings and not isinstance(setting_value, (list, tuple)):
                    raise TypeError(f'{setting} must be a list or a tuple.')
                setattr(self, setting, setting_value)
                self._explicit_settings.add(setting)

        if not self.SECRET_KEY:
            raise AttributeError('The SECRET_KEY setting must not be empty.')

    def __repr__(self):
        return f'<{type(self).__name__} {self.SETTINGS_MODULE}>'


settings = LazySettings()



from importlib import import_module

from falcon_core.utils import flatten


class Route:
    def __init__(self, uri_template, resource, **kwargs):
        if not uri_template or uri_template.startswith('/') and '//' not in uri_template:
            self.uri_template = uri_template
            self.resource = resource
            self.kwargs = kwargs
        else:
            raise ValueError(f'Route uri_template \'{uri_template}\' is not valid.')


def route(*args, **kwargs):
    ri = Route(*args, **kwargs)
    if isinstance(ri.resource, list):
        ri.resource = flatten(ri.resource)
        for r in ri.resource:
            r.uri_template = f'{ri.uri_template}{r.uri_template}'
        return ri.resource
    return ri


def include(routes):
    if isinstance(routes, str):
        return import_module(routes).routes
    return routes



from waitress import serve

from falcon_core.management.base import BaseCommand
from falcon_core.config import settings
from falcon_core.utils import import_object


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('address', action='store', nargs='?', default='127.0.0.1:8000')

    def handle(self, *args, **kwargs):
        address = kwargs.pop('address')
        host, port = address.split(':')
        serve(import_object(settings.WSGI_APPLICATION), host=host, port=port)
